
# Copyright 2018, 2019 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

import socket
import select
import time
import json
import sys
import threading
from netq_lib.common.enums import ConfKw
from netq_lib.orm import HOSTNAME
from netq_lib.common.netq_pq import QueueElem
import psutil
import os
from collections import defaultdict

import grpc
from netq_agent.interface.netq_opta_stream_pb2 import (
    NetqAgentMessage,
)

from netq_agent.interface.netq_opta_stream_pb2_grpc import (
    NetqOPTAStreamServiceStub
)

class AGENTSTATUS():
    SUCCESS = 0
    FAILURE = 1


class ChannelState():
    IDLE = 1
    CONNECTING = 2
    CONNECTED = 3
    DISCONNECTED = 4
    SHUTDOWN = 5

class TransMode():
    PARTIAL_TRANSFORM = 1
    FULL_TRANSFORM = 0
    FIRST_SEGMENT = 2
    MID_SEGMENT = 3
    LAST_SEGMENT = 4


RPC_CONNECTIVITY_MAPPING = {}
RPC_CONNECTIVITY_MAPPING[grpc.ChannelConnectivity.IDLE] = ChannelState.IDLE
RPC_CONNECTIVITY_MAPPING[grpc.ChannelConnectivity.CONNECTING] = ChannelState.DISCONNECTED
RPC_CONNECTIVITY_MAPPING[grpc.ChannelConnectivity.READY] = ChannelState.CONNECTED
RPC_CONNECTIVITY_MAPPING[grpc.ChannelConnectivity.TRANSIENT_FAILURE] = ChannelState.DISCONNECTED
RPC_CONNECTIVITY_MAPPING[grpc.ChannelConnectivity.SHUTDOWN] = ChannelState.IDLE


class NetQAgentGrpcClient(object):

    def __init__(self, conf, logger, pq, backpressure_queue):
        self._conf = conf
        self._channel = None
        self._logger = logger
        self.channel_state = ChannelState.IDLE
        self._pq = pq
        self._bp_pq = backpressure_queue
        self.seq_no = 0
        self._max_opta_retry = 3
        self._grpc_backoff_time = 0

        self.__MAX_GRPC_BACKOFF = 30
        self.__chunk_count = {}
        self.__total_chunk_count = 0
        self.__opta_grps_reset_count = 0
        self.__util_stats = {}
        self.__pq_util_stats = defaultdict(int)

        # Rate Control */
        self._wait_time = 0.0
        self._last_transmission_time = 0.0
        self._elapsed_time = 0.0
        self._min_interval = 0.2
        self._max_message_per_transaction = 500
        self._max_msg_threshold = 100
        self._wakeup_time = 10
        self._wait_for_ready_timeout = 60

        try:
            self._num_msg_per_packet = self._conf['netq-agent']['num_msgs_per_packet']
        except Exception:
            self._num_msg_per_packet = 30

        self._logger.info("Number of messagees per packet {}".format(self._num_msg_per_packet))


    def get_sequence_number(self):
        self.seq_no = self.seq_no + 1
        return self.seq_no


    def callback_on_exception(self, exception, traceback):

        self._logger.error(traceback)
        self._logger.error(exception)


    def _run():
        pass

    @property
    def channel_state(self):
        return self._channel_state


    @channel_state.setter
    def channel_state(self, new_state):
        if new_state in RPC_CONNECTIVITY_MAPPING:
            new_state = RPC_CONNECTIVITY_MAPPING[new_state]
        self._logger.debug("New RPC Channel State {}".format(new_state))
        self._channel_state = new_state


    def _get_agent_connect_id(self, verbose=False):
        id = None
        try:
            opta_url = self._conf['netq-agent'][ConfKw.SERVER]
            opta_port = self._conf['netq-agent'].get(ConfKw.PORT, ConfKw.DEFAULT_OPTA_PORT)
            s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            s.connect((opta_url, opta_port))
            id = s.getsockname()[0]
        except Exception:
            self._logger.error("Failed to connect to OPTA")

        if verbose:
            if id:
                self._logger.debug("OPTA Discovery Agent Connect IP {}".format(id))
            else:
                self._logger.debug("OPTA Discovery No Connect IP determined")
        return id


    def get_agent_id(self):
        if not self.is_valid_hostname(HOSTNAME):
            return None
        return HOSTNAME


    def is_valid_hostname(self, hostname):
        if hostname.lower() in ["cumulus"]:
            return False
        return True


    def full_state_message_builder(self, msg_type, segment, seq_no, json_out):

        netq_message = NetqAgentMessage(aid=HOSTNAME,
                                        message_type=msg_type,
                                        seqno=seq_no,
                                        ts=long(time.time() * 1000),
                                        trans_mode=segment,
                                        type=0,
                                        message=json.dumps(json_out).encode('utf-8'))
        return netq_message

    def message_builder(self, msg_type, fullstate, json_out):

        full_state = TransMode.FULL_TRANSFORM if fullstate else TransMode.PARTIAL_TRANSFORM

        if len(json_out) > 0:
            event_time = long(json_out[0].get('timestamp', 0) * 1000)
        else:
            event_time = long(time.time() * 1000)

        netq_message = NetqAgentMessage(aid=HOSTNAME,
                                        message_type=msg_type,
                                        seqno=self.get_sequence_number(),
                                        ts=event_time,
                                        trans_mode=full_state,
                                        type=0,
                                        message=json.dumps(json_out).encode('utf-8'))
        return netq_message

    def get_full_state_message(self, msg):

        num_msg_per_packet = self._num_msg_per_packet
        if 'priority_qid' not in msg.keys() or 'message_type' not in msg.keys():
            self._logger.error('Invalid full state message')

        pq_idx = msg.get('priority_qid')
        queue_len = self._pq.len(pq_idx)
        num_segments = int(((queue_len - 1) / num_msg_per_packet) + 1)
        counter = msg.get('counter')
        self._logger.debug("pq_idx = {} queue_len = {} num_segments = {}".format(pq_idx, queue_len, num_segments))

        msg_data = []
        num_msgs = 0

        seq_no = self.get_sequence_number()
        segment_idx = 1

        for priority, message_type, fullstate, msg in self._pq.get_buffer(pq_idx, counter):
            msg_data.append(msg.data)
            num_msgs += 1
            if (num_msgs == num_msg_per_packet):

                self.__chunk_count[message_type] = self.__chunk_count.get(message_type, 0) + 1
                self.__total_chunk_count = self.__total_chunk_count + 1

                if (segment_idx == num_segments):
                    rpc_msg = self.full_state_message_builder(message_type, TransMode.LAST_SEGMENT, seq_no, msg_data)
                elif(segment_idx == 1):
                    rpc_msg = self.full_state_message_builder(message_type, TransMode.FIRST_SEGMENT, seq_no, msg_data)
                else:
                    rpc_msg = self.full_state_message_builder(message_type, TransMode.MID_SEGMENT, seq_no, msg_data)
                msg_data = []
                num_msgs = 0
                segment_idx += 1
                yield rpc_msg

        if len(msg_data) > 0:
            rpc_msg = self.full_state_message_builder(message_type, TransMode.LAST_SEGMENT, seq_no, msg_data)
            yield rpc_msg

    def _get_wait_time(self):
        self._elapsed_time = time.time() - self._last_transmission_time
        return self._min_interval - self._elapsed_time

    def grpc_init_rate_control(self):
        self._wait_time = self._get_wait_time()
        self._logger.debug("elapsed time {} wait time {} min interval {} queue size {}".format(self._elapsed_time, self._wait_time, self._min_interval, self._pq.len(1)))
        while self._wait_time > 0:
            q_len = self._pq.len_all()
            if q_len > self._max_msg_threshold:
                break
            else:
                time.sleep(self._wakeup_time)
            self._wait_time = self._get_wait_time()

    def _send_wakeup_message(self):
        message = {}
        message['message_type'] = 'NETQ_RATE_CONTROL'
        message['priority_qid'] = 1
        priority = 1
        try:
            queue_elem = QueueElem(priority, "NETQ_RATE_CONTROL", True, message)
            self._pq.putq(queue_elem)
        except Exception:
            return False

    def process_msg_data(self, msg_type, json_out):

        if msg_type == "NETQ_RATE_CONTROL":
            return True

        elif msg_type == "AgentStats":

            pq_util = {'average_pq_util': float(self.__pq_util_stats.get('sum_pq_util')) / self.__pq_util_stats.get('count'),
                       'max_pq_util': self.__pq_util_stats.get('max_pq_util', 0),
                       'min_pq_util': self.__pq_util_stats.get("min_pq_util", 0)}

            json_out['pq_util'] = pq_util
            self.__pq_util_stats = defaultdict(int)

            json_out['message_count']['total_chunk_count'] = self.__total_chunk_count
            self.__total_chunk_count = 0
            json_out['chunk_count'] = self.__chunk_count
            self.__chunk_count = {}
            json_out['opta_grps_reset_count'] = self.__opta_grps_reset_count
            self.__opta_grps_reset_count = 0

            grpc_util = {'grpc_pid': self.__util_stats.get('grpc_pid', os.getpid()),
                         'max_cpu_util': self.__util_stats.get('max_cpu_util', 0.0),
                         'min_cpu_util': self.__util_stats.get('min_cpu_util', 0.0),
                         'avg_cpu_util': self.__util_stats.get('sum_cpu_util', 0.0) / self.__util_stats.get('count', 1),
                         'max_mem_util': self.__util_stats.get('max_mem_util', 0.0),
                         'min_mem_util': self.__util_stats.get('min_mem_util', 0.0),
                         'avg_mem_util': self.__util_stats.get('sum_mem_util', 0.0) / self.__util_stats.get('count', 1)}

            json_out['grpc_util'] = grpc_util
            self.__util_stats = {}

            return False

        else:
            return False

    def get_stream(self):

        pq = self._pq
        fd_set = pq.get_reader_fds()
        my_run_id = 1


        try:
            (epoll_input, [], []) = select.select(fd_set, [], [], 10)
            if not len(epoll_input):
                return
            self.grpc_init_rate_control()
            msg_cnt = 0

            for queue_idx in range(0, self._pq.max_num_queue):
                for priority, message_type, fullstate, msg in pq.get_buffer(queue_idx):

                    ignore_msg_type = self.process_msg_data(message_type, msg.data)

                    if ignore_msg_type:
                        continue

                    if message_type == 'FULL_STATE_END':
                        full_gen = self.get_full_state_message(msg.data)
                        for segment in full_gen:
                            yield segment
                            msg_cnt += 1
                    elif isinstance(msg.data, list):
                        rpc_msg = self.message_builder(message_type, fullstate, msg.data)
                        yield rpc_msg
                    else:
                        rpc_msg = self.message_builder(message_type, fullstate, [msg.data])
                        yield rpc_msg
                    msg_cnt += 1
                    if msg_cnt >= self._max_message_per_transaction:
                        self._logger.debug("Reached max messages per transaction")
                        self._send_wakeup_message()
                        break
            self._last_transmission_time = time.time()

        except Exception as e:
            self._logger.error("GRPC %d StreamMessage failed %s", my_run_id, e)
            self._last_transmission_time = time.time()
            return


    def _get_opta_addr(self):

        serv_addr = "{}:{}".format(self._conf['netq-agent'][ConfKw.SERVER],
                                   self._conf['netq-agent'].get(ConfKw.PORT, ConfKw.DEFAULT_OPTA_PORT))
        return serv_addr


    def monitor_channel(self, state):
        self._logger.info("Channel State {}".format(state))
        self.channel_state = state


    def wait_for_ready(self):
        try:
            grpc.channel_ready_future(self._channel).result(timeout=self._wait_for_ready_timeout)
            self._logger.info("Channel connection successful")
            return AGENTSTATUS.SUCCESS
        except grpc.FutureTimeoutError:
            self._logger.info("Channel Failed to connect")
            return AGENTSTATUS.FAILURE


    def reestablish_channel(self):

        serv_addr = self._get_opta_addr()
        self.__opta_grps_reset_count += 1

        self._logger.info("Reestablishing OPTA Connection {}".format(serv_addr))

        if self._channel:
            return self.wait_for_ready()

        try:
            self._channel = grpc.insecure_channel(serv_addr)
        except Exception as e:
            self._logger.error(
                "Unable to establish connection with OPTA {}".format(e))
            return AGENTSTATUS.FAILURE

        if self._channel:
            self._channel.subscribe(self.monitor_channel, try_to_connect=True)
            self.wait_for_ready()
        else:
            self._logger.error("Unable to connect to OPTA")
            return AGENTSTATUS.FAILURE

        if self.channel_state != ChannelState.CONNECTED:
            return AGENTSTATUS.FAILURE

        return AGENTSTATUS.SUCCESS

    def _close(self):
        try:
            if self._channel:
                self._logger.debug("Closing channel {}".format(self._channel))
                self._channel.unsubscribe(self.monitor_channel)
                self._channel.close()
            self._channel = None
        except Exception as e:
            self._logger.error("Channel Close Error {}".format(e))
            self._channel = None

    def run_contiguous_stream(self):

        opta_stub = NetqOPTAStreamServiceStub(self._channel)
        metadata = [('aid', HOSTNAME)]

        while True:
            try:
                res, call = opta_stub.StreamMessage.with_call(self.get_stream(), metadata=metadata)
                self._logger.debug("RPC Completed {} status {}".format(res, call.code()))
                if call.code() != grpc.StatusCode.OK:
                    self._logger.error("Agent Stream Failed {}".format(call.code()))
                    self._close()
                    return AGENTSTATUS.FAILURE
            except Exception as e:
                self._logger.error("Agent Stream Failed {}".format(e))
                self._close()
                return AGENTSTATUS.FAILURE

    def run_grpc_stats_collection(self):
        while True:
            try:
                self.__pq_util_stats = self._pq.total_pq_usage(self.__pq_util_stats)

                agent_process = psutil.Process(os.getpid())
                self.__util_stats['grpc_pid'] = os.getpid()

                agent_cpu_usage = agent_process.cpu_percent(interval=0.1)
                total_mem = psutil.virtual_memory().total
                agent_mem_usage = float((agent_process.memory_info().vms * 100)) / total_mem

                self.__util_stats['count'] = self.__util_stats.get('count', 0) + 1

                if agent_cpu_usage >= self.__util_stats.get('max_cpu_util', 0.0):
                    self.__util_stats['max_cpu_util'] = agent_cpu_usage

                if agent_cpu_usage <= self.__util_stats.get('min_cpu_util', sys.maxint):
                    self.__util_stats['min_cpu_util'] = agent_cpu_usage


                self.__util_stats['sum_cpu_util'] = self.__util_stats.get('sum_cpu_util', 0.0) + agent_cpu_usage

                if agent_mem_usage >= self.__util_stats.get('max_mem_util', 0.0):
                    self.__util_stats['max_mem_util'] = agent_mem_usage

                if agent_mem_usage <= self.__util_stats.get('min_mem_util', 0.0):
                    self.__util_stats['min_mem_util'] = agent_mem_usage

                self.__util_stats['sum_mem_util'] = self.__util_stats.get('sum_mem_util', 0.0) + agent_mem_usage


                time.sleep(15)

            except Exception as e:
                self._logger.error("Agent Stream Failed {}".format(e))
                return AGENTSTATUS.FAILURE

    def send_pause_state(self, state):
        self._logger.debug("Sending pause state {}".format(state))
        try:
            q_elem = QueueElem(0, 'bp', state, 'bp')
            self._bp_pq.putq(q_elem)
            if state is not True:
                for queue_idx in range(0, self._pq.max_num_queue):
                    self._pq.empty_pq(queue_idx)
            else:
                self._reset_grpc_backoff()
            self._logger.debug("Agent Connectivity State {}".format(state))

        except Exception as ex:
            self._logger.error('GRPC backpressure putq failed %s', ex)


    def _get_grpc_backoff_time(self):
        self._logger.debug("Current Back of time {}".format(self._grpc_backoff_time))
        if self._grpc_backoff_time >= self.__MAX_GRPC_BACKOFF:
            return self.__MAX_GRPC_BACKOFF
        else:
            self._grpc_backoff_time = 1 + self._grpc_backoff_time * 2
            return self._grpc_backoff_time


    def _reset_grpc_backoff(self):
        self._logger.debug("Resetting GRPC backoff time")
        self._grpc_backoff_time = 0


    def start_stream(self, verbose=False):

        retryCnt = 1

        stats_thread = threading.Thread(target=self.run_grpc_stats_collection, args=())
        stats_thread.start()

        while True:

            self._logger.info("Agent Stream started")

            if AGENTSTATUS.SUCCESS != self.reestablish_channel():
                self._logger.debug("Failed to establish Connection with OPTA, retrying with backoff")
                if (retryCnt >= self._max_opta_retry):
                    self.send_pause_state(False)
                    retryCnt = 0
                else:
                    retryCnt = retryCnt + 1
                timeVal = self._get_grpc_backoff_time()
                time.sleep(timeVal)
                continue

            self._logger.info("Starting Stream")
            self.send_pause_state(True)

            try:
                t1 = threading.Thread(target=self.run_contiguous_stream, args=())
                #stream_response = opta_stub.StreamMessage(self.get_stream())
                t1.start()
                t1.join()
                self._logger.debug("Thread procesisng complete")
            except Exception as e:
                self._logger.error("OPTA Stream Unknown GRPC Exception {}".format(e))
                return AGENTSTATUS.FAILURE

            self._logger.debug("Stream Completed")
